;;;; -*- Mode: Lisp; Syntax: ANSI-Common-Lisp; Base: 10 -*-
;;;; *************************************************************************
;;;; FILE IDENTIFICATION
;;;;
;;;; Name:          mop.lisp
;;;; Purpose:       Metaobject Protocol Interface
;;;; Programmer:    Kevin M. Rosenberg
;;;; Date Started:  Apr 2000
;;;;
;;;; This metaclass as functions to classes to allow display
;;;; in Text, HTML, and XML formats. This includes hyperlinking
;;;; capability and sub-objects.
;;;;
;;;; $Id: mop.lisp,v 1.71 2003/05/14 05:36:22 kevin Exp $
;;;;
;;;; This file is Copyright (c) 2000-2003 by Kevin M. Rosenberg
;;;; *************************************************************************
 
(in-package :hyperobject)

(eval-when (:compile-toplevel :execute)
  (declaim (optimize (speed 2) (safety 2) (compilation-speed 0) (debug 2))))

;; Main class

(defclass hyperobject-class (standard-class)
  ( ;; slots initialized in defclass
   (user-name :initarg :user-name :type string :initform nil
	      :accessor user-name
	      :documentation "User name for class")
   (user-name-plural :initarg :user-name-plural :type string :initform nil
		     :accessor user-name-plural
		     :documentation "Plural user name for class")
   (default-print-slots :initarg :default-print-slots :type list :initform nil
			:accessor default-print-slots
			:documentation "Defaults slots for a view")
   (description :initarg :description :initform nil
		:accessor description
		:documentation "Class description")
   (version :initarg :version :initform nil
	    :accessor version
	    :documentation "Version number for class")
   (sql-name :initarg :sql-name :initform nil)

   ;;; The remainder of these fields are calculated one time
   ;;; in finalize-inheritence.
   
   (subobjects :initform nil :accessor subobjects
	       :documentation
	       "List of fields that contain a list of subobjects objects.")
   (hyperlinks :type list :initform nil :accessor hyperlinks
	       :documentation "List of fields that have hyperlinks")
   (direct-rules :type list :initform nil :initarg :direct-rules
		 :accessor direct-rules
		 :documentation "List of rules to fire on slot changes.")
   (class-id :type integer :initform nil
	     :accessor class-id
	     :documentation "Unique ID for the class")
   (default-view :initform nil :initarg :default-view :accessor default-view
		 :documentation "The default view for a class")

   ;; SQL commands
   (create-table-cmd :initform nil :reader create-table-cmd)
   (create-indices-cmds :initform nil :reader create-index-cmds)
   (drop-table-cmd :initform nil :reader drop-table-cmd)

   (views :type list :initform nil :initarg :views :accessor views
	  :documentation "List of views")
   (rules :type list :initform nil :initarg :rules :accessor rules
	  :documentation "List of rules")
   )
  (:documentation "Metaclass for Markup Language classes."))

(defclass subobject ()
  ((name-class :type symbol :initform nil :initarg :name-class :reader name-class)
   (name-slot :type symbol :initform nil :initarg :name-slot :reader name-slot)
   (lookup :type (or function symbol) :initform nil :initarg :lookup :reader lookup)
   (lookup-keys :type list :initform nil :initarg :lookup-keys
		:reader lookup-keys))
  (:documentation "Contains subobject information"))


(defmethod print-object ((obj subobject) (s stream))
  (print-unreadable-object (obj s :type t :identity t)
    (format s "~S" (name obj))))

(defclass hyperlink ()
  ((name :type symbol :initform nil :initarg :name :reader name)
   (lookup
    ;; The type specifier seems to break sbcl
    :type (or function symbol)
    ;;    :type t
    :initform nil :initarg :lookup :reader lookup)
   (link-parameters :type list :initform nil :initarg :link-parameters
		    :reader link-parameters)))

(defmethod print-object ((obj hyperlink) (s stream))
  (print-unreadable-object (obj s :type t :identity t)
    (format s "~S" (name obj))))

#+(or cmu scl sbcl)
(defmethod validate-superclass ((class hyperobject-class) (superclass standard-class))
  t)

(defmethod finalize-inheritance :after ((cl hyperobject-class))
  (init-hyperobject-class cl)
  )

(eval-when (:compile-toplevel :load-toplevel :execute)
  (when (>= (length (generic-function-lambda-list
		     (ensure-generic-function
		      'compute-effective-slot-definition)))
	    3)
    (pushnew :ho-normal-cesd cl:*features*))
  
    (when (>= (length (generic-function-lambda-list
		       (ensure-generic-function
			'direct-slot-definition-class)))
	    3)
      (pushnew :ho-normal-dsdc cl:*features*))
    
    (when (>= (length (generic-function-lambda-list
		       (ensure-generic-function
			'effective-slot-definition-class)))
	      3)
      (pushnew :ho-normal-esdc cl:*features*)))

;; Slot definitions
(defmethod direct-slot-definition-class ((cl hyperobject-class)
					 #+ho-normal-dsdc &rest iargs)
  (find-class 'hyperobject-dsd))

(defmethod effective-slot-definition-class ((cl hyperobject-class) 
					    #+ho-normal-esdc &rest iargs)
  (find-class 'hyperobject-esd))


;;; Slot definitions

(eval-when (:compile-toplevel :load-toplevel :execute)
  (defmacro process-class-option (slot-name &optional required)
    #+lispworks
    `(defmethod clos:process-a-class-option ((class hyperobject-class)
					     (name (eql ,slot-name))
					     value)
      (when (and ,required (null value))
	(error "hyperobject class slot ~A must have a value" name))
      (list name `',value))
    #+(or allegro sbcl cmu scl)
    (declare (ignore slot-name required))
    )

  (defmacro process-slot-option (slot-name)
    #+lispworks
    `(defmethod clos:process-a-slot-option ((class hyperobject-class)
					    (option (eql ,slot-name))
					    value
					    already-processed-options
					    slot)
      (list* option `',value already-processed-options))
    #-lispworks
    (declare (ignore slot-name))
    )
  
  (dolist (option *class-options*)
    (eval `(process-class-option ,option)))
  (dolist (option *slot-options*)
    (eval `(process-slot-option ,option)))

  (eval
   `(defclass hyperobject-dsd (standard-direct-slot-definition)
     (,@(mapcar #'(lambda (x)
		    `(,(intern (symbol-name x))
		      :initform nil))
		*slot-options-no-initarg*)
      ,@(mapcar #'(lambda (x)
		    `(,(intern (symbol-name x))
		      :initarg
		      ,(intern (symbol-name x) (symbol-name :keyword))
		      :initform nil
		      :accessor
		      ,(intern (concatenate 'string
					    (symbol-name :dsd-)
					    (symbol-name x)))))
		*slot-options*))))
  (eval
   `(defclass hyperobject-esd (standard-effective-slot-definition)
     (,@(mapcar #'(lambda (x)
		    `(,(intern (symbol-name x))
		      :initarg
		      ,(intern (symbol-name x) (symbol-name :keyword))
		      :initform nil
		      :accessor
		      ,(intern (concatenate 'string
					    (symbol-name :esd-)
					    (symbol-name x)))))
		(append *slot-options* *slot-options-no-initarg*)))))
  ) ;; eval-when

(defun intern-in-keyword (obj)
  (cond
    ((null obj)
     nil)
    ((eq t obj)
     t)
    ((atom obj)
     (intern (symbol-name obj) (find-package 'keyword)))
    ((consp obj)
     (cons (intern-in-keyword (car obj) ) (intern-in-keyword (cdr obj))))
    (t
     obj)))

(defun canonicalize-value-type (vt)
  (typecase vt
    (atom
     (ensure-keyword vt))
    (cons
     (cons (ensure-keyword (car vt)) (cdr vt)))
    (t
     t)))

#+ignore
(defmethod compute-effective-slot-definition :around ((cl hyperobject-class) #+ho-normal-cesd name dsds)
  #+allegro (declare (ignore name))
  (let* ((dsd (car dsds))
	 (value-type (canonicalize-value-type (slot-value dsd 'value-type))))
    (multiple-value-bind (sql-type length) (value-type-to-sql-type value-type)
      (setf (slot-value dsd 'sql-type) sql-type)
      (setf (slot-value dsd 'type) (value-type-to-lisp-type value-type))
      (let ((ia (compute-effective-slot-definition-initargs cl #+lispworks name dsds)))
	(apply
	 #'make-instance 'hyperobject-esd 
	 :value-type value-type
	 :sql-type sql-type
	 :length length
	 :print-formatter (slot-value dsd 'print-formatter)
	 :subobject (slot-value dsd 'subobject)
	 :hyperlink (slot-value dsd 'hyperlink)
	 :hyperlink-parameters (slot-value dsd 'hyperlink-parameters)
	 :description (slot-value dsd 'description)
	 :user-name (slot-value dsd 'user-name)
	 :user-name-plural (slot-value dsd 'user-name-plural)
	 :index (slot-value dsd 'index)
	 :value-constraint (slot-value dsd 'value-constraint)
	 :null-allowed (slot-value dsd 'null-allowed)
	 ia)))))

(defmethod compute-effective-slot-definition :around ((cl hyperobject-class) #+ho-normal-cesd name dsds)
  #+ho-normal-cesd (declare (ignore name))
  (let* ((esd (call-next-method))
	 (dsd (car dsds))
	 (value-type (canonicalize-value-type (slot-value dsd 'value-type))))
    (multiple-value-bind (sql-type length) (value-type-to-sql-type value-type)
      (setf (slot-value esd 'sql-type) sql-type)
      (setf (slot-value esd 'length) length)
      (setf (slot-value esd 'type) (value-type-to-lisp-type value-type))
      (setf (slot-value esd 'value-type) value-type)
      (dolist (name '(print-formatter subobject hyperlink hyperlink-parameters
		      description value-constraint index null-allowed user-name))
	(setf (slot-value esd name) (slot-value dsd name)))
      esd)))


#+ho-normal-cesd
(setq cl:*features* (delete :ho-normal-cesd cl:*features*))
#+ho-normal-dsdc
(setq cl:*features* (delete :ho-normal-dsdc cl:*features*))
#+ho-normal-esdc
(setq cl:*features* (delete :ho-normal-esdc cl:*features*))

(defun lisp-type-is-a-string (type)
  (or (eq type 'string)
      (and (listp type) (some #'(lambda (x) (eq x 'string)) type))))

(defun value-type-to-lisp-type (value-type)
  (case (if (atom value-type)
	    value-type
	    (car value-type))
    ((:string :cdata :varchar :char)
     '(or null string))
    (:character
     '(or null character))
    (:fixnum
     '(or null fixnum))
    (:boolean
     '(or null boolean))
    (:integer
     '(or null integer))
    ((:float :single-float)
     '(or null single-float))
    (:double-float
     '(or null double-float))
    (otherwise
     t)))

(defun value-type-to-sql-type (value-type)
  "Return two values, the sql type and field length."
  (let ((type (if (atom value-type)
		  value-type
		  (car value-type)))
	(length (when (consp value-type)
		  (cadr value-type))))
    (values
     (case type
       ((:string :cdata)
	:string)
       ((:fixnum :integer)
	:integer)
       (:boolean
	:boolean)
       ((:float :single-float)
	:single-float)
       (:double-float
	:double-float)
       (otherwise
	:text))
     length)))

;;;; Class initialization function

;; defines a slot-unbound method for class and slot-name, fills
;; the slot by calling reader function with the slot values of
;; the instance's reader-keys
(defmacro def-lazy-reader (class slot-name reader &rest reader-keys)
  (let* ((the-slot-name (gensym))
	 (the-class (gensym))
	 (the-instance (gensym))
	 (keys '()))
    (dolist (key reader-keys)
      (push (list 'slot-value the-instance (list 'quote key)) keys))
    (setq keys (nreverse keys))
    `(defmethod slot-unbound (,the-class (,the-instance ,class)
			      (,the-slot-name (eql ',slot-name)))
      (declare (ignore ,the-class))
      (setf (slot-value ,the-instance ,the-slot-name) (,reader ,@keys)))))


#+lispworks
(defun intern-eql-specializer (slot)
  `(eql ,slot))

#+(or sbcl cmu lispworks)
(defun ensure-lazy-reader (class-name slot-name reader &rest reader-keys)
  (let ((keys nil)
	(gf (ensure-generic-function 'slot-unbound)))
    (dolist (key reader-keys)
      (push (list 'slot-value 'the-instance (list 'quote key)) keys))
    (setq keys (nreverse keys))
    (multiple-value-bind (method-lambda init-args-values)
	(make-method-lambda
	 gf
	 (class-prototype (generic-function-method-class gf))
	 #-lispworks
	 `(lambda (the-class the-instance the-slot-name)
	   (declare (ignore the-class))
	   (setf (slot-value the-instance the-slot-name) (,reader ,@keys)))
	 #+lispworks
	 '(the-class the-instance the-slot-name)
	 #+lispworks
	 nil
	 #+lispworks
	 `(setf (slot-value the-instance the-slot-name) (,reader ,@keys))
	 nil)
      (add-method gf
		  (apply
		   #'make-instance (generic-function-method-class gf)
		   ':specializers (list (class-of (find-class class-name))
					(find-class class-name)
					(intern-eql-specializer slot-name))
		   ':lambda-list '(the-class the-instance the-slot-name)
		   ':function (compile nil method-lambda)
		   init-args-values)))))

(defun finalize-subobjects (cl)
  "Process class subobjects slot"
  (setf (subobjects cl)
	(let ((subobjects '()))
	  (dolist (slot (class-slots cl))
	    (let-when
	     (subobj-def (esd-subobject slot))
	     (let ((subobject
		    (make-instance 'subobject
				   :name-class (class-name cl)
				   :name-slot (slot-definition-name slot)
				   :lookup (if (atom subobj-def)
					       subobj-def
					       (car subobj-def))
				   :lookup-keys (if (atom subobj-def)
						    nil
						    (cdr subobj-def)))))
	       (unless (eq (lookup subobject) t)
		 #-(or sbcl cmu lispworks)
		 (eval
		  `(hyperobject::def-lazy-reader ,(name-class subobject)
		    ,(name-slot subobject) ,(lookup subobject)
		    ,@(lookup-keys subobject)))
		 #+(or sbcl cmu lispworks)
		 (apply #'ensure-lazy-reader 
			(name-class subobject) (name-slot subobject)
			(lookup subobject) (lookup-keys subobject)))
	       (push subobject subobjects))))
	  ;; sbcl/cmu reverse class-slots compared to the defclass form
	  ;; so re-reverse on cmu/sbcl
	  #+(or cmu sbcl) subobjects
	  #-(or cmu sbcl) (nreverse subobjects)
	  )))


(defun finalize-class-slots (cl)
  "Make sure all class slots have an expected value"
  (unless (user-name cl)
    (setf (user-name cl) (format nil "~:(~A~)" (class-name cl))))
  
  (setf (user-name-plural cl)
	(if (and (consp (user-name cl)) (cadr (user-name cl)))
	    (cadr (user-name cl))
	    (format nil "~A~P" (if (consp (user-name cl))
				   (car (user-name cl))
				   (user-name cl))
		    2)))

  (dolist (name '(user-name description))
    (awhen (slot-value cl name)
	   (setf (slot-value cl name)
		 (etypecase (slot-value cl name)
		   (cons (car it))
		   ((or string symbol) it))))))

(defun finalize-documentation (cl)
  "Calculate class documentation slot"
  (let ((*print-circle* nil))
    (setf (documentation (class-name cl) 'class)
	  (format nil "Hyperobject~A~A~A~A"
		  (aif (user-name cl)
		       (format nil ": ~A" it ""))
		  (aif (description cl)
		       (format nil "~%Class description: ~A" it) "")
		  (aif (subobjects cl)
		       (format nil "~%Subobjects:~{ ~A~}" (mapcar #'name-slot it)) "")
		  (aif (default-print-slots cl)
		       (format nil "~%Default print slots:~{ ~A~}" it) "")
		  ))))

(defun finalize-hyperlinks (cl)
  (let ((hyperlinks '()))
    (dolist (esd (class-slots cl))
      (awhen (slot-value esd 'hyperlink)
	     (push
	      (make-instance 'hyperlink
			     :name (slot-definition-name esd)
			     :lookup it
			     :link-parameters (slot-value esd 'hyperlink-parameters))
	      hyperlinks)))
    ;; cmu/sbcl reverse class-slots compared to the defclass form
    ;; hyperlinks is already reversed from the dolist/push loop, so re-reverse on sbcl/cmu
    #-(or cmu sbcl) (setq hyperlinks (nreverse hyperlinks))
    (setf (slot-value cl 'hyperlinks) hyperlinks)))

(defun init-hyperobject-class (cl)
  "Initialize a hyperobject class. Calculates all class slots"
  (finalize-subobjects cl)
  (finalize-views cl)
  (finalize-hyperlinks cl)
  (finalize-sql cl)
  (finalize-rules cl)
  (finalize-class-slots cl)
  (finalize-documentation cl))


;;;; *************************************************************************
;;;;  Metaclass Slot Accessors
;;;; *************************************************************************

(defun find-slot-by-name (cl name)
  (find name (class-slots cl) :key #'slot-definition-name))

(defun hyperobject-class-user-name (obj)
  (user-name (class-of obj)))

(defun hyperobject-class-user-name-plural (obj)
  (user-name-plural (class-of obj)))

(defun hyperobject-class-subobjects (obj)
  (subobjects (class-of obj)))

(defun hyperobject-class-hyperlinks (obj)
  (hyperlinks (class-of obj)))

(defun hyperobject-class-slots (obj)
  ;; cmucl/sbcl reverse class-slots
  #+(or cmu sbcl) (reverse (class-slots (class-of obj)))
  #-(or cmu sbcl) (class-slots (class-of obj)))

